@@ -18,15 +18,15 @@
 #include <sys/types.h>
 #include <sys/mount.h>
 #include <linux/types.h>
-#include <linux/fs.h>
 #include <linux/hdreg.h>
 #include <linux/major.h>
+#include <asm/byteorder.h>
 
 #include "hdparm.h"
 
 extern const char *minor_str[];
 
-#define VERSION "v5.2"
+#define VERSION "v5.3"
 
 #undef DO_FLUSHCACHE		/* under construction: force cache flush on -W0 */
 
@@ -777,6 +777,7 @@ void process_dev (char *devname)
 	}
 	if (!quiet)
 		printf("\n%s:\n", devname);
+
 	if (set_readahead) {
 		if (get_readahead)
 			printf(" setting fs readahead to %ld\n", readahead);
@@ -810,8 +811,12 @@ void process_dev (char *devname)
 		if (noisy_piomode) {
 			if (piomode == 255)
 				printf(" attempting to auto-tune PIO mode\n");
-			else
+			else if (piomode < 100)
 				printf(" attempting to set PIO mode to %d\n", piomode);
+			else if (piomode < 200)
+				printf(" attempting to set MDMA mode to %d\n", (piomode-100));
+			else
+				printf(" attempting to set UDMA mode to %d\n", (piomode-200));
 		}
 		if (ioctl(fd, HDIO_SET_PIO_MODE, piomode)) 
 			perror(" HDIO_SET_PIO_MODE failed");
@@ -1278,17 +1283,21 @@ void process_dev (char *devname)
 #ifdef HDIO_SET_32BIT
 	if (get_IDentity) {
 		unsigned char args[4+512] = {WIN_IDENTIFY,0,0,1,};
+		unsigned i;
 		no_scsi();
 		no_xt();
-		if (!(ioctl(fd, HDIO_DRIVE_CMD, &args))) {
-			identify((void *)&args[4], NULL);
-		} else {
+		if (ioctl(fd, HDIO_DRIVE_CMD, &args)) {
 			args[0] = WIN_PIDENTIFY;
-			if (!(ioctl(fd, HDIO_DRIVE_CMD, &args))) {
-				identify((void *)&args[4], NULL);
-			} else
+			if (ioctl(fd, HDIO_DRIVE_CMD, &args)) {
 				perror(" HDIO_DRIVE_CMD(identify) failed");
+				goto identify_abort;
+			}
 		}
+		for(i=0; i<(sizeof args)/2; i+=2) {
+		    __le16_to_cpus((__u16 *)(&args[i]));
+		}    
+		identify((void *)&args[4], NULL);
+identify_abort:
 	}
 #endif
 #ifdef HDIO_SET_BUSSTATE
@@ -1338,6 +1347,17 @@ void process_dev (char *devname)
 		}
 	}
  #endif /* HDIO_GET_ACOUSTIC */
+
+	//
+	// Try to adjust buffersize based on device size
+	//
+	if (do_ctimings || do_timings) {
+		if (ioctl(fd, BLKGETSIZE, &parm))
+			perror(" BLKGETSIZE failed");
+		else if ( (parm * 512/(1024 * 1024)) < timing_MB)
+			timing_MB = parm * 512/(1024 * 1024);
+	}
+
 	if (do_ctimings)
 		time_cache (fd);
 	if (do_timings)
@@ -1473,9 +1493,9 @@ static int fromhex (unsigned char c)
 
 static int identify_from_stdin (void)
 {
-	unsigned char *buf = malloc(1600), *b = buf;
-	unsigned short *w = (unsigned short *)buf;
-	int i, count = read(0, buf, 1280);
+	unsigned short buf[800];
+	unsigned char *b = (unsigned char *)buf;
+	int i, count = read(0, b, 1280);
 
 	if (count != 1280) {
 		fprintf(stderr, "read(1280 bytes) failed (rc=%d)", count);
@@ -1483,7 +1503,8 @@ static int identify_from_stdin (void)
 		exit(errno);
 	}
 	for (i = 0; count >= 4; ++i) {
-		*w++ = (fromhex(*b++)<<12) | (fromhex(*b++)<<8) | (fromhex(*b++)<<4) | (fromhex(*b++));
+		buf[i] = (fromhex(*b++) << 12) | (fromhex(*b++) << 8) | (fromhex(*b++) << 4) | fromhex(*b++);
+		__le16_to_cpus((__u16 *)(&buf[i]));
 		++b;
 		count -= 5;
 	}
